/***************************************************************************
                          ToyLoad.c

    Contains the code for processing .lod files.
                             -------------------
    begin                : Sat Jan 25 2003
    copyright            : (C) 2003 by Tyler Montbriand
    student #            : 200200370
    class                : CS330
    email                : tsm@accesscomm.ca
 ***************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "ToySegment.h"
#include "ToySegmentList.h"
#include "ToyProcess.h"
#include "ToyMemType.h"
#include "ToyMem.h"
#include "ToyMem_Internal.h"
#include "toyparse.h"
#include "ToyLoad.h"

#define SEG_CODE 0
#define SEG_DATA 1
#define SEG_HEAP 2
#define SEG_STACK 3

const char  *DataType_Magic[]       = {  "8585",NULL           };
const char  *DataType_Integer[]     = {  "%u","#%u",NULL       };
const char  *DataType_String[]      = {  "\"%s",NULL           };
const char  *DataType_Instruction[] = {  "RUN","HALT",NULL     };
const char **DataType_List[]        = {  DataType_Magic,
                                                DataType_Integer,
                                                DataType_String,
                                                DataType_Instruction,
                                                NULL                  };

/*    Loader file sequence:
        Magic #
        # segments(max 4)
        Segment sizes
        Data(fills each segment)   */

char **GetExplodedLine(char *buf, int len, FILE *fp,int *n);

ErrMessage_t ToyMM_Load(ToyMM *tm,int process,const char *filename,int addr)
{
  int n           =-1;
  int numSegments =-1;
  int segSizes[4] ={-1,-1,-1,-1};
  char buf[160]   ={'\0'};
  char **strList  =NULL;  
  FILE *fileIn    =NULL;
  /*  Can't open nonexistent files,
      nor use nonexistent toy-mm's  */
  if((tm==NULL)||(filename==NULL))          return(ERR_UNKNOWN);

  if((process<0)||(process>=MAX_PROCESSES))
  {
    return(ERR_UNKNOWN);
  }

  if(tm->processList[process]!=NULL)
  {
    return(ERR_PROCESSEXISTS);
  }
  
  fileIn=fopen(filename,"r");
  if(fileIn==NULL)  return(ERR_NOFILE); /*  Couldn't open file*/

  /* Get the first line.  It should be "8585" */
  strList=GetExplodedLine(buf,160,fileIn,&n);
  BAIL_IF(strList==NULL,fclose(fileIn),ERR_INVALIDFILE);
  BAIL_IF(DetermineType(strList[0],DataType_List)!=DATA_MAGIC,fclose(fileIn); FreeStrList(strList),ERR_INVALIDFILE);
  FreeStrList(strList);

  /*  Get the second line.  Should be the # of
      segments, i.e. "4"                         */
  strList=GetExplodedLine(buf,160,fileIn,&n);
  BAIL_IF(n!=1,fclose(fileIn); FreeStrList(strList),ERR_INVALIDFILE);
  BAIL_IF(DetermineType(strList[0],DataType_List)!=DATA_INTEGER,fclose(fileIn); FreeStrList(strList),ERR_INVALIDFILE);
  
  /*  OK, read the # into numSegments */
  sscanf(strList[0],"%u",&numSegments);
  BAIL_IF((numSegments<=0)||(numSegments>4),FreeStrList(strList),ERR_INVALIDFILE);
  FreeStrList(strList);

  /*  Get the segment sizes.  Should be like "1 3 8 4"  */
  strList=GetExplodedLine(buf,160,fileIn,&n);
  BAIL_IF((strList==NULL),fclose(fileIn); ,ERR_INVALIDFILE);  
  BAIL_IF(n!=numSegments,fclose(fileIn);FreeStrList(strList),ERR_INVALIDINSTRUCTION);
  for(n=0; n<numSegments; n++)
  {
    BAIL_IF(DetermineType(strList[n],DataType_List)!=DATA_INTEGER,fclose(fileIn); FreeStrList(strList),ERR_INVALIDFILE);
    sscanf(strList[n],"%u",&segSizes[n]);
  }
  FreeStrList(strList);

  /*  Allocate the process. */
  if(TMM_AllocateProcessAt(tm,process,addr,numSegments,segSizes)!=ERR_NONE)
  {
    fclose(fileIn); 
    return(ERR_INSUFFICIENTMEM);
  }

  { /*  OK, now we're cookin'.  Load stuff into memory! */
    int i=-1;
    int segnum    =-1;       /*  Current segment       */
    int numStrings=-1;      /*  # of strings in list  */
    int segpos    =-1;      /*  Position in segment   */
    int abspos    =-1;      /*  absolute memory position  */
    SegList *list = NULL;   /*  Current segment       */

    segpos=0;
    segnum=0;
    list=tm->processList[process]->segments;  /*Get the list of segments*/
    list=TMM_SeekSegListHead(list);

    /*  Get the next line of input  */
    strList=GetExplodedLine(buf,160,fileIn,&numStrings);
    while(strList!=NULL)
    {
      for(i=0; i<numStrings; i++)
      {
        int type=-1,subtype=-1;

        /*  If we've reached the end of this segment,
            advance to the next one.                */
        if(segpos>=list->seg.len)
        {
          segpos=0;
          list=list->next;
          segnum++;
        }

        /*  If there is no next segment, we're out of memory. */
        if(list==NULL)
        {
          /*  Out of segment space! */
          TMM_TerminateProcess(tm,process);
          FreeStrList(strList);
          return(ERR_DATAOVERFLOW);
        }

        /* Calculate our position in memory */
        abspos=segpos+list->seg.pos;
        /*  Determine type and subtype  */
        type=DetermineType(strList[i],DataType_List);
        subtype=DetermineSubType(type,strList[i],DataType_List);
        
        switch(segnum)
        {
        case SEG_CODE:
          if(type!=DATA_INSTRUCTION)
          {
            fclose(fileIn);
            FreeStrList(strList);
            TMM_TerminateProcess(tm,process);
            return(ERR_INVALIDINSTRUCTION);
          }
          break;

        case SEG_DATA:
          if((type!=DATA_INTEGER)&&(type!=DATA_STRING))
          {
            fclose(fileIn);
            FreeStrList(strList);
            TMM_TerminateProcess(tm,process);
            return(ERR_INVALIDDATA);
          }
          break;

        case SEG_STACK:
        case SEG_HEAP:
          fclose(fileIn);
          FreeStrList(strList);
          TMM_TerminateProcess(tm,process);
          return(ERR_DATAOVERFLOW);
        }

        if(type==DATA_ERROR)  /*  If the parser doesn't recognize it  */
        {
          FreeStrList(strList);
          fclose(fileIn);
          return(ERR_INVALIDDATA);
        }

        tm->mem[abspos].type=type;

        switch(type)
        {
          case DATA_INTEGER:
          {
            int val;
            sscanf(strList[i]+1,"%d",&val);
            tm->mem[abspos].integer.value=val;
          }
          break;
        case DATA_INSTRUCTION:
          tm->mem[abspos].integer.value=subtype;
          break;

        case DATA_STRING:
          tm->mem[abspos].string.value[0]='\0';
          if(strlen(strList[i])>MAX_STRLEN)
          {
            FreeStrList(strList);
            fclose(fileIn);
            return(ERR_INVALIDDATA);
            break;
          }
          
          strcpy(tm->mem[abspos].string.value,strList[i]+1);

          break;
        }

        segpos++;   
      }
      FreeStrList(strList); /*  Free list of strings  */
      strList=GetExplodedLine(buf,160,fileIn,&numStrings);  /*  Get next line */
    }
  }

  /*  Wow, we made it!  */
  fclose(fileIn);
  
  return(ERR_NONE);  /*  Return success  */
}

char **GetExplodedLine(char *buf, int len, FILE *fp,int *n)
{
  char **strList=NULL;
  buf[0]='\0';            /*  Destroy prev contents     */

  if(!feof(fp))
  {
    fgets(buf,len,fp);    /*  Get text                  */
    strList=Explode(buf); /*  Explode text into tokens  */
  }

  if(n!=NULL)  /*  Count # of tokens */
  {
    if(strList==NULL) ((*n)=-1);  /*  -1 means no tokens  */
    else
    {
     for((*n)=0; strList[(*n)]!=NULL; (*n)++);
    }
  }              
  return(strList);
}
