/***************************************************************************
                          ToyParseCmd.c  -  Code to parse Toy Memory
    Manager command files.
                             -------------------
    begin                : Sat Jan 25 2003
    copyright            : (C) 2003 by Tyler Montbriand
    student #            : 200200370
    class                : CS330
    email                : tsm@accesscomm.ca 
***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <stdio.h>
#include <string.h>

#include "ToyMem.h"
#include "toyparse.h"
#include "ToyMemType.h"
#include "ToyLoad.h"
#include "ToyParseCmd.h"

/*  These lists of strings are carefully crafted to
    match their enumerated types.  So searching the
    tree of strings will return the proper value. */

/* List of argument forms for processes */
const char *ArgType_Process[]={"P%u",NULL};
/* List of argument forms for strings */
const char *ArgType_String[]={"\"%s",NULL};
/* List of argument forms for addresses */
const char *ArgType_Address[]={"@%u",NULL};
/* List of argument forms for integers */
const char *ArgType_Integer[]={"#%u",NULL};
/* List of all commands */
const char *ArgType_Command[]={ "LoadFile","PrintProcessTable",
                                "PrintProcessMemory","STORE",
                                "TerminateProcess","PrintAllMemory",
                                "END",NULL                          };

/* Array of arrays of pointers to acceptable kinds of arguments. */
const char **Arguments[]={  ArgType_Process,ArgType_Command,ArgType_Address,
                            ArgType_Integer,ArgType_String,NULL               };


int ReadFromType(ArgType_e type, const char *str, void *buffer)
{
  switch(type)
  {
  case TYPE_STRING:
    if(strlen(str+1)>MAX_STRLEN)  return -1;
    strcpy(buffer,str+1);
    return(0);

    //All the same kind of integer - unsigned 32 bit
  case TYPE_PROCESS:
  case TYPE_INTEGER:
  case TYPE_ADDRESS:
    sscanf(str+1,"%u",(unsigned int *)buffer);
    return(0);

  default:
    return(-1);
  }
}

/* Parse a line from a config file. */
ErrMessage_t TMM_ParseLine(const char *strIn, ToyMM *, int line);

/*  Parses a command-file.  Returns 0 on success, -1 on error.  */
int TMM_ParseFile(const char *fname, ToyMM *tm)
{
  int line=0;
  char buf[160]={'\0'};
  ErrMessage_t error=ERR_UNKNOWN;
  FILE *fp=NULL;  

  fp=fopen(fname,"r");        /*  Open specified file for textmode reading  */
  if(fp==NULL)  return(-1);   /*  If we didn't get the file, return error   */

  fgets(buf,160,fp);    /*  Read first line  */
  
  /*  Ignore blank lines  */
  if(strlen(buf)>1) error=TMM_ParseLine(buf,tm,++line);    /*  Parse line  */
  else              error=ERR_NONE;
  
  do  /*  Loop until we've read it all  */
  {
    if(error!=ERR_NONE)     /*  if the line caused an error */
    {
      if(error==ERR_UNKNOWN)
      {
        /*  Unknown errors have no string in the error table.  */
        printf(ERR_PREFIX"Unknown error.\n");
      }
      else
      {
        /*  But other errors do.  */
        printf(ERR_PREFIX"%s\n",ErrMessages[error]);
      }
      if(error==ERR_PARSEEND) /*  END command, stop parsing */
      {
        fclose(fp);
        return(-1);
      }
    }
    buf[0]='\0';        /*  Empty the buffer  */    
    fgets(buf,160,fp);  /*  Read line         */
    
    /*  Ignore blank lines  */
    if(strlen(buf)>1) error=TMM_ParseLine(buf,tm,++line);    /*  Parse line  */
    else              error=ERR_NONE;
  }
  while(!feof(fp));     /*  Loop until end-of-file  */

  fclose(fp);           /*  Close the file  */

  return(0);  /*  Return success. */
}

/* Parses the given line of input.  Returns on success, -1 on error. */
ErrMessage_t TMM_ParseLine(const char *strIn, ToyMM *tm, int line)
{
  int process=-1,n=-1;
  char **strList=NULL;
  CommandType_e command=COMMAND_ERROR;

  strList=Explode(strIn);
  BAIL_IF(strList==NULL,,-1);

  /*  Print the input, and Count # of arguments  */

  if(strList[0]!=NULL)
  {
    printf(LINE_NUMBER_FORMAT" ",line);
    printf("%s",strList[0]);
    for(n=1; strList[n]!=NULL; n++) printf(" %s",strList[n]);
    printf("\n");
  }

  if(n<2)
  {  /*At least 2 arguments are required for ALL operations*/
    FreeStrList(strList);
    return(ERR_INVALIDINPUT);
  }
  else if(DetermineType(strList[0],Arguments)!=TYPE_PROCESS)
  {  /*  All lines must start with a process # */
    FreeStrList(strList);
    return(ERR_INVALIDINPUT);
  }

  ReadFromType(TYPE_PROCESS,strList[0],&process);

  if((process>=MAX_PROCESSES)||(process<0))
  {   /*  Can't command nonexistent processes */    
    FreeStrList(strList);
    return(ERR_INVALIDINPUT);
  }
  else if(DetermineType(strList[1],Arguments)!=TYPE_COMMAND)
  {  /*  Second argument MUST be a command */
    FreeStrList(strList);
    return(ERR_INVALIDINPUT);
  }

  command=DetermineSubType(TYPE_COMMAND,strList[1],Arguments);
  
  /*  Shouldn't happen, but test just in case */
  BAIL_IF(command==COMMAND_ERROR,FreeStrList(strList),ERR_INVALIDINPUT);
  switch(command)
  {
    char string[100];
    int addr;
    int val;

    case COMMAND_END:
      return(ERR_PARSEEND);
    
    case COMMAND_PRINTALLMEMORY:
      /*  Should be only 2 arguments - process #, and command */
      BAIL_IF(n!=2,FreeStrList(strList),ERR_INVALIDINPUT);
      /*  Only process 0 can do this  */
      BAIL_IF(process!=0,FreeStrList(strList),ERR_NOPRIVILEGE);
      return(TMM_PrintAllMemory(tm));

    case COMMAND_LOADFILE:
      /*  LoadFile requires 4 arguments */
      BAIL_IF(n!=4,FreeStrList(strList),ERR_INVALIDINPUT);
      /*  3rd argument must be a memory address */      
      BAIL_IF(DetermineType(strList[2],Arguments)!=TYPE_ADDRESS,FreeStrList(strList),ERR_INVALIDINPUT);
      /*  4th argument must be a string filename  */
      BAIL_IF(DetermineType(strList[3],Arguments)!=TYPE_STRING,FreeStrList(strList),ERR_INVALIDINPUT);

      /*  Read data from these types  */
      ReadFromType(TYPE_ADDRESS,strList[2],&addr);
      ReadFromType(TYPE_STRING,strList[3],string);
      FreeStrList(strList);

      /*  Out-of-range means a syntax error, not a memory one for some reason.  */
      if(addr>=tm->memSize) return(ERR_INVALIDINPUT);      
      else                  return(ToyMM_Load(tm,process,string,addr));

    case COMMAND_PRINTPROCESSTABLE:
      /*  Exactly 2 parameters - process ID, and the  command */
      BAIL_IF(n!=2,FreeStrList(strList),ERR_INVALIDINPUT);
      FreeStrList(strList);
      return(TMM_PrintProcessTable(tm,process));
      break;

    case COMMAND_PRINTPROCESSMEMORY:
      /*  Exactly 2 paramegers - process ID, and command  */
      BAIL_IF(n!=2,FreeStrList(strList),ERR_INVALIDINPUT);
      FreeStrList(strList);
      
      return(TMM_PrintProcessMemory(tm,process));
      break;

    case COMMAND_STORE:
      /*  Exactly 4 parameters - process ID, command, address, and data */
      BAIL_IF(n!=4, FreeStrList(strList),ERR_INVALIDINPUT);
      /*  3rd parameter MUST be an address  */
      BAIL_IF(DetermineType(strList[2],Arguments)!=TYPE_ADDRESS,FreeStrList(strList),ERR_INVALIDINPUT);
      ReadFromType(TYPE_ADDRESS,strList[2],&addr);
      
      /*  Out-of-range means a syntax error, not a memory one for some reason.  */
      if(addr>=tm->memSize) return(ERR_INVALIDINPUT);

      /*  4th parameter can be either an integer or a string  */
      if( (DetermineType(strList[3],Arguments)==TYPE_INTEGER)&&
          (ReadFromType(TYPE_INTEGER,strList[3],&val)>=0))
      {
        FreeStrList(strList);
        return(TMM_StoreInteger(tm,process,addr,val));
      }
      else if(  (DetermineType(strList[3],Arguments)==TYPE_STRING)&&
                (ReadFromType(TYPE_STRING,strList[3],string)>=0))
      {
        FreeStrList(strList);
        return(TMM_StoreString(tm,process,addr,string));
      }
      else
      {
        FreeStrList(strList);
        return(ERR_INVALIDINPUT);
      }
      break;

    case COMMAND_TERMINATEPROCESS:
      /*  Must be precisely 2 parameters - process ID and command */
      BAIL_IF(n!=2,FreeStrList(strList),ERR_INVALIDINPUT);

      return(TMM_TerminateProcess(tm,process));
      break;

    default:
      /*  This should not happen under any circumstances. */
      printf("Email me @ tsm@accesscomm.ca and tell me what you did please!\n");
      BAIL_IF(1,FreeStrList(strList),ERR_UNKNOWN);
      break;
  }

  FreeStrList(strList);
  return(0);
}
